@@ -10,7 +10,7 @@
 
 use anyhow::{bail, Context, Result};
 use std::collections::{HashMap, HashSet};
-use std::fmt::Write;
+use std::fmt::{self, Write};
 use std::mem;
 use std::path::Path;
 use wasmparser::*;
@@ -52,12 +52,20 @@ struct ModuleState {
     module: u32,
     instance: u32,
     memory: u32,
-    event: u32,
+    tag: u32,
     global: u32,
     table: u32,
+    label: u32,
     types: Vec<Option<FuncType>>,
-    names: HashMap<u32, Naming>,
-    local_names: HashMap<u32, HashMap<u32, Naming>>,
+    function_names: HashMap<u32, Naming>,
+    local_names: HashMap<(u32, u32), Naming>,
+    label_names: HashMap<(u32, u32), Naming>,
+    type_names: HashMap<u32, Naming>,
+    table_names: HashMap<u32, Naming>,
+    memory_names: HashMap<u32, Naming>,
+    global_names: HashMap<u32, Naming>,
+    element_names: HashMap<u32, Naming>,
+    data_names: HashMap<u32, Naming>,
     module_name: Option<Naming>,
     implicit_instances_seen: HashSet<String>,
 }
@@ -195,7 +203,7 @@ impl Printer {
                 }
                 Payload::TableSection(s) => self.print_tables(s)?,
                 Payload::MemorySection(s) => self.print_memories(s)?,
-                Payload::EventSection(s) => self.print_events(s)?,
+                Payload::TagSection(s) => self.print_tags(s)?,
                 Payload::GlobalSection(s) => self.print_globals(s)?,
                 Payload::ExportSection(s) => self.print_exports(s)?,
                 Payload::StartSection { func, .. } => {
@@ -251,38 +259,51 @@ impl Printer {
     }
 
     fn register_names(&mut self, names: NameSectionReader<'_>) -> Result<()> {
+        fn name_map(into: &mut HashMap<u32, Naming>, names: NameMap<'_>) -> Result<()> {
+            let mut used = HashSet::new();
+            let mut map = names.get_map()?;
+            for _ in 0..map.get_count() {
+                let naming = map.read()?;
+                into.insert(naming.index, Naming::new(naming.name, &mut used));
+            }
+            Ok(())
+        }
+
+        fn indirect_name_map(
+            into: &mut HashMap<(u32, u32), Naming>,
+            names: IndirectNameMap<'_>,
+        ) -> Result<()> {
+            let mut outer_map = names.get_indirect_map()?;
+            for _ in 0..outer_map.get_indirect_count() {
+                let mut used = HashSet::new();
+                let outer = outer_map.read()?;
+                let mut inner_map = outer.get_map()?;
+                for _ in 0..inner_map.get_count() {
+                    let inner = inner_map.read()?;
+                    into.insert(
+                        (outer.indirect_index, inner.index),
+                        Naming::new(inner.name, &mut used),
+                    );
+                }
+            }
+            Ok(())
+        }
+
         for section in names {
             match section? {
                 Name::Module(n) => {
                     let name = Naming::new(n.get_name()?, &mut HashSet::new());
                     self.state.module_name = Some(name);
                 }
-                Name::Function(n) => {
-                    let mut names = HashSet::new();
-                    let mut map = n.get_map()?;
-                    for _ in 0..map.get_count() {
-                        let name = map.read()?;
-                        self.state
-                            .names
-                            .insert(name.index, Naming::new(name.name, &mut names));
-                    }
-                }
-                Name::Local(n) => {
-                    let mut reader = n.get_function_local_reader()?;
-                    for _ in 0..reader.get_count() {
-                        let mut names = HashSet::new();
-                        let local_name = reader.read()?;
-                        let mut map = local_name.get_map()?;
-                        let mut local_map = HashMap::new();
-                        for _ in 0..map.get_count() {
-                            let name = map.read()?;
-                            local_map.insert(name.index, Naming::new(name.name, &mut names));
-                        }
-                        self.state
-                            .local_names
-                            .insert(local_name.func_index, local_map);
-                    }
-                }
+                Name::Function(n) => name_map(&mut self.state.function_names, n)?,
+                Name::Local(n) => indirect_name_map(&mut self.state.local_names, n)?,
+                Name::Label(n) => indirect_name_map(&mut self.state.label_names, n)?,
+                Name::Type(n) => name_map(&mut self.state.type_names, n)?,
+                Name::Table(n) => name_map(&mut self.state.table_names, n)?,
+                Name::Memory(n) => name_map(&mut self.state.memory_names, n)?,
+                Name::Global(n) => name_map(&mut self.state.global_names, n)?,
+                Name::Element(n) => name_map(&mut self.state.element_names, n)?,
+                Name::Data(n) => name_map(&mut self.state.data_names, n)?,
                 Name::Unknown { .. } => (),
             }
         }
@@ -292,8 +313,9 @@ impl Printer {
     fn print_types(&mut self, parser: TypeSectionReader<'_>) -> Result<()> {
         for ty in parser {
             self.newline();
-            self.start_group("type");
-            write!(self.result, " (;{};) ", self.state.types.len())?;
+            self.start_group("type ");
+            self.print_cur_type_name()?;
+            self.result.push_str(" ");
             let ty = match ty? {
                 TypeDef::Func(ty) => {
                     self.start_group("func");
@@ -344,13 +366,13 @@ impl Printer {
         names_for: Option<u32>,
     ) -> Result<Option<u32>> {
         if always_print_type {
-            write!(self.result, " (type {})", idx)?;
+            self.print_type_ref(idx)?;
         }
         let ty = match self.state.types.get(idx as usize) {
             Some(Some(ty)) => ty.clone(),
             Some(None) => {
                 if !always_print_type {
-                    write!(self.result, " (type {})", idx)?;
+                    self.print_type_ref(idx)?;
                 }
                 return Ok(None);
             }
@@ -372,9 +394,7 @@ impl Printer {
         // a new one if that's the case with a named parameter.
         for (i, param) in ty.params.iter().enumerate() {
             let local_names = &self.state.local_names;
-            let name = names_for
-                .and_then(|n| local_names.get(&n))
-                .and_then(|n| n.get(&(i as u32)));
+            let name = names_for.and_then(|n| local_names.get(&(n, i as u32)));
             params.start_local(name, &mut self.result);
             self.print_valtype(*param)?;
             params.end_local(&mut self.result);
@@ -439,7 +459,7 @@ impl Printer {
                 ImportSectionEntryType::Instance(_) => self.state.instance += 1,
                 ImportSectionEntryType::Table(_) => self.state.table += 1,
                 ImportSectionEntryType::Memory(_) => self.state.memory += 1,
-                ImportSectionEntryType::Event(_) => self.state.event += 1,
+                ImportSectionEntryType::Tag(_) => self.state.tag += 1,
                 ImportSectionEntryType::Global(_) => self.state.global += 1,
             }
         }
@@ -467,30 +487,27 @@ impl Printer {
                 self.start_group("func");
                 if index {
                     self.result.push_str(" ");
-                    match self.state.names.get(&self.state.func) {
-                        Some(name) => name.write(&mut self.result),
-                        None => write!(self.result, "(;{};)", self.state.func)?,
-                    }
+                    self.print_cur_func_name()?;
                 }
-                write!(self.result, " (type {})", f)?;
+                self.print_type_ref(*f)?;
             }
             Module(f) => {
-                self.start_group("module ");
+                self.start_group("module");
                 if index {
-                    write!(self.result, "(;{};) ", self.state.module)?;
+                    write!(self.result, " (;{};)", self.state.module)?;
                 }
-                write!(self.result, "(type {})", f)?;
+                self.print_type_ref(*f)?;
             }
             Instance(f) => {
-                self.start_group("instance ");
+                self.start_group("instance");
                 if index {
-                    write!(self.result, "(;{};) ", self.state.instance)?;
+                    write!(self.result, " (;{};)", self.state.instance)?;
                 }
-                write!(self.result, "(type {})", f)?;
+                self.print_type_ref(*f)?;
             }
             Table(f) => self.print_table_type(&f, index)?,
             Memory(f) => self.print_memory_type(&f, index)?,
-            Event(f) => self.print_event_type(&f, index)?,
+            Tag(f) => self.print_tag_type(&f, index)?,
             Global(f) => self.print_global_type(&f, index)?,
         }
         self.end_group();
@@ -500,9 +517,10 @@ impl Printer {
     fn print_table_type(&mut self, ty: &TableType, index: bool) -> Result<()> {
         self.start_group("table ");
         if index {
-            write!(self.result, "(;{};) ", self.state.table)?;
+            self.print_cur_table_name()?;
+            self.result.push_str(" ");
         }
-        self.print_limits(&ty.limits)?;
+        self.print_limits(ty.initial, ty.maximum)?;
         self.result.push_str(" ");
         self.print_valtype(ty.element_type)?;
         Ok(())
@@ -511,40 +529,34 @@ impl Printer {
     fn print_memory_type(&mut self, ty: &MemoryType, index: bool) -> Result<()> {
         self.start_group("memory ");
         if index {
-            write!(self.result, "(;{};) ", self.state.memory)?;
+            self.print_cur_memory_name()?;
+            self.result.push_str(" ");
         }
-        match ty {
-            MemoryType::M32 { limits, shared } => {
-                self.print_limits(limits)?;
-                if *shared {
-                    self.result.push_str(" shared");
-                }
-            }
-            MemoryType::M64 { limits, shared } => {
-                write!(self.result, "i64 {}", limits.initial)?;
-                if let Some(max) = limits.maximum {
-                    write!(self.result, " {}", max)?;
-                }
-                if *shared {
-                    self.result.push_str(" shared");
-                }
-            }
+        if ty.memory64 {
+            self.result.push_str("i64 ");
+        }
+        self.print_limits(ty.initial, ty.maximum)?;
+        if ty.shared {
+            self.result.push_str(" shared");
         }
         Ok(())
     }
 
-    fn print_event_type(&mut self, ty: &EventType, index: bool) -> Result<()> {
-        self.start_group("event ");
+    fn print_tag_type(&mut self, ty: &TagType, index: bool) -> Result<()> {
+        self.start_group("tag ");
         if index {
-            write!(self.result, "(;{};)", self.state.event)?;
+            write!(self.result, "(;{};)", self.state.tag)?;
         }
         self.print_functype_idx(ty.type_index, true, None)?;
         Ok(())
     }
 
-    fn print_limits(&mut self, limits: &ResizableLimits) -> Result<()> {
-        write!(self.result, "{}", limits.initial)?;
-        if let Some(max) = limits.maximum {
+    fn print_limits<T>(&mut self, initial: T, maximum: Option<T>) -> Result<()>
+    where
+        T: fmt::Display,
+    {
+        write!(self.result, "{}", initial)?;
+        if let Some(max) = maximum {
             write!(self.result, " {}", max)?;
         }
         Ok(())
@@ -553,7 +565,8 @@ impl Printer {
     fn print_global_type(&mut self, ty: &GlobalType, index: bool) -> Result<()> {
         self.start_group("global ");
         if index {
-            write!(self.result, "(;{};) ", self.state.global)?;
+            self.print_cur_global_name()?;
+            self.result.push_str(" ");
         }
         if ty.mutable {
             self.result.push_str("(mut ");
@@ -587,13 +600,13 @@ impl Printer {
         Ok(())
     }
 
-    fn print_events(&mut self, parser: EventSectionReader<'_>) -> Result<()> {
-        for exn in parser {
-            let exn = exn?;
+    fn print_tags(&mut self, parser: TagSectionReader<'_>) -> Result<()> {
+        for tag in parser {
+            let tag = tag?;
             self.newline();
-            self.print_event_type(&exn, true)?;
+            self.print_tag_type(&tag, true)?;
             self.end_group();
-            self.state.event += 1;
+            self.state.tag += 1;
         }
         Ok(())
     }
@@ -620,10 +633,7 @@ impl Printer {
             let ty = funcs.read()?;
             self.newline();
             self.start_group("func ");
-            match self.state.names.get(&self.state.func) {
-                Some(name) => name.write(&mut self.result),
-                None => write!(self.result, "(;{};)", self.state.func)?,
-            }
+            self.print_cur_func_name()?;
             let params = self
                 .print_functype_idx(ty, true, Some(self.state.func))?
                 .unwrap_or(0);
@@ -648,8 +658,7 @@ impl Printer {
                     let name = self
                         .state
                         .local_names
-                        .get(&self.state.func)
-                        .and_then(|m| m.get(&(params + local_idx)));
+                        .get(&(self.state.func, params + local_idx));
                     locals.start_local(name, &mut self.result);
                     self.print_valtype(ty)?;
                     locals.end_local(&mut self.result);
@@ -658,8 +667,10 @@ impl Printer {
             }
             locals.finish(&mut self.result);
 
+            self.state.label = 0;
             let nesting_start = self.nesting;
             let mut reader = body.get_operators_reader()?;
+            reader.allow_memarg64(true);
             while !reader.eof() {
                 let operator = reader.read()?;
                 match operator {
@@ -716,8 +727,8 @@ impl Printer {
 
     fn print_operator(&mut self, op: &Operator<'_>, nesting_start: u32) -> Result<()> {
         use Operator::*;
-        let cur_label = self.nesting - nesting_start;
-        let label = |relative: u32| match cur_label.checked_sub(relative) {
+        let cur_depth = self.nesting - nesting_start;
+        let label = |relative: u32| match cur_depth.checked_sub(relative) {
             Some(i) => format!("@{}", i),
             None => format!(" INVALID "),
         };
@@ -726,24 +737,20 @@ impl Printer {
             Unreachable => self.result.push_str("unreachable"),
             Block { ty } => {
                 self.result.push_str("block");
-                self.print_blockty(ty)?;
-                write!(self.result, "  ;; label = @{}", cur_label)?;
+                self.print_blockty(ty, cur_depth)?;
             }
             Loop { ty } => {
                 self.result.push_str("loop");
-                self.print_blockty(ty)?;
-                write!(self.result, "  ;; label = @{}", cur_label)?;
+                self.print_blockty(ty, cur_depth)?;
             }
             If { ty } => {
                 self.result.push_str("if");
-                self.print_blockty(ty)?;
-                write!(self.result, "  ;; label = @{}", cur_label)?;
+                self.print_blockty(ty, cur_depth)?;
             }
             Else => self.result.push_str("else"),
             Try { ty } => {
                 self.result.push_str("try");
-                self.print_blockty(ty)?;
-                write!(self.result, "  ;; label = @{}", cur_label)?;
+                self.print_blockty(ty, cur_depth)?;
             }
             Catch { index } => {
                 write!(self.result, "catch {}", index)?;
@@ -792,9 +799,10 @@ impl Printer {
             CallIndirect { table_index, index } => {
                 self.result.push_str("call_indirect");
                 if *table_index != 0 {
-                    write!(self.result, " {}", table_index)?;
+                    self.result.push_str(" ");
+                    self.print_table_idx(*table_index)?;
                 }
-                write!(self.result, " (type {})", index)?;
+                self.print_type_ref(*index)?;
             }
             ReturnCall { function_index } => {
                 self.result.push_str("return_call ");
@@ -803,9 +811,10 @@ impl Printer {
             ReturnCallIndirect { table_index, index } => {
                 self.result.push_str("return_call_indirect");
                 if *table_index != 0 {
-                    write!(self.result, " {}", table_index)?;
+                    self.result.push_str(" ");
+                    self.print_table_idx(*table_index)?;
                 }
-                write!(self.result, " (type {})", index)?;
+                self.print_type_ref(*index)?;
             }
 
             Delegate { relative_depth } => {
@@ -834,10 +843,12 @@ impl Printer {
             }
 
             GlobalGet { global_index } => {
-                write!(self.result, "global.get {}", global_index)?;
+                self.result.push_str("global.get ");
+                self.print_global_idx(*global_index)?;
             }
             GlobalSet { global_index } => {
-                write!(self.result, "global.set {}", global_index)?;
+                self.result.push_str("global.set ");
+                self.print_global_idx(*global_index)?;
             }
 
             I32Load { memarg } => self.mem_instr("i32.load", memarg, 4)?,
@@ -866,9 +877,15 @@ impl Printer {
             I64Store32 { memarg } => self.mem_instr("i64.store32", memarg, 4)?,
 
             MemorySize { mem: 0, .. } => self.result.push_str("memory.size"),
-            MemorySize { mem, .. } => write!(self.result, "memory.size {}", mem)?,
+            MemorySize { mem, .. } => {
+                self.result.push_str("memory.size ");
+                self.print_memory_idx(*mem)?;
+            }
             MemoryGrow { mem: 0, .. } => self.result.push_str("memory.grow"),
-            MemoryGrow { mem, .. } => write!(self.result, "memory.grow {}", mem)?,
+            MemoryGrow { mem, .. } => {
+                self.result.push_str("memory.grow ");
+                self.print_memory_idx(*mem)?;
+            }
 
             I32Const { value } => write!(self.result, "i32.const {}", value)?,
             I64Const { value } => write!(self.result, "i64.const {}", value)?,
@@ -1040,37 +1057,75 @@ impl Printer {
             I64TruncSatF64S => self.result.push_str("i64.trunc_sat_f64_s"),
             I64TruncSatF64U => self.result.push_str("i64.trunc_sat_f64_u"),
 
-            MemoryInit { segment, mem: 0 } => write!(self.result, "memory.init {}", segment)?,
-            MemoryInit { segment, mem } => write!(self.result, "memory.init {} {}", segment, mem)?,
-            DataDrop { segment } => write!(self.result, "data.drop {}", segment)?,
-            MemoryCopy { src: 0, dst: 0 } => self.result.push_str("memory.copy"),
-            MemoryCopy { src, dst } => write!(self.result, "memory.copy {} {}", dst, src)?,
-            MemoryFill { mem: 0 } => self.result.push_str("memory.fill"),
-            MemoryFill { mem } => write!(self.result, "memory.fill {}", mem)?,
-
-            TableInit { table, segment } => {
-                if *table == 0 {
-                    write!(self.result, "table.init {}", segment)?
-                } else {
-                    write!(self.result, "table.init {} {}", table, segment)?
+            MemoryInit { segment, mem } => {
+                self.result.push_str("memory.init ");
+                self.print_data_idx(*segment)?;
+                if *mem != 0 {
+                    self.result.push_str(" ");
+                    self.print_memory_idx(*mem)?;
                 }
             }
-            ElemDrop { segment } => write!(self.result, "elem.drop {}", segment)?,
+            DataDrop { segment } => {
+                self.result.push_str("data.drop ");
+                self.print_data_idx(*segment)?;
+            }
+            MemoryCopy { src: 0, dst: 0 } => self.result.push_str("memory.copy"),
+            MemoryCopy { src, dst } => {
+                self.result.push_str("memory.copy ");
+                self.print_memory_idx(*dst)?;
+                self.result.push_str(" ");
+                self.print_memory_idx(*src)?;
+            }
+            MemoryFill { mem: 0 } => self.result.push_str("memory.fill"),
+            MemoryFill { mem } => {
+                self.result.push_str("memory.fill ");
+                self.print_memory_idx(*mem)?;
+            }
+
+            TableInit { table, segment } => {
+                self.result.push_str("table.init ");
+                if *table != 0 {
+                    self.print_table_idx(*table)?;
+                    self.result.push_str(" ");
+                }
+                self.print_elem_idx(*segment)?;
+            }
+            ElemDrop { segment } => {
+                self.result.push_str("elem.drop ");
+                self.print_elem_idx(*segment)?;
+            }
             TableCopy {
                 dst_table,
                 src_table,
             } => {
-                if *dst_table == *src_table && *src_table == 0 {
-                    self.result.push_str("table.copy");
-                } else {
-                    write!(self.result, "table.copy {} {}", dst_table, src_table)?
+                self.result.push_str("table.copy");
+                if *dst_table != *src_table || *src_table != 0 {
+                    self.result.push_str(" ");
+                    self.print_table_idx(*dst_table)?;
+                    self.result.push_str(" ");
+                    self.print_table_idx(*src_table)?;
                 }
             }
-            TableGet { table } => write!(self.result, "table.get {}", table)?,
-            TableSet { table } => write!(self.result, "table.set {}", table)?,
-            TableGrow { table } => write!(self.result, "table.grow {}", table)?,
-            TableSize { table } => write!(self.result, "table.size {}", table)?,
-            TableFill { table } => write!(self.result, "table.fill {}", table)?,
+            TableGet { table } => {
+                self.result.push_str("table.get ");
+                self.print_table_idx(*table)?;
+            }
+            TableSet { table } => {
+                self.result.push_str("table.set ");
+                self.print_table_idx(*table)?;
+            }
+            TableGrow { table } => {
+                self.result.push_str("table.grow ");
+                self.print_table_idx(*table)?;
+            }
+            TableSize { table } => {
+                self.result.push_str("table.size ");
+                self.print_table_idx(*table)?;
+            }
+            TableFill { table } => {
+                self.result.push_str("table.fill ");
+                self.print_table_idx(*table)?;
+            }
 
             MemoryAtomicNotify { memarg } => self.mem_instr("memory.atomic.notify", memarg, 4)?,
             MemoryAtomicWait32 { memarg } => self.mem_instr("memory.atomic.wait32", memarg, 4)?,
@@ -1484,7 +1539,9 @@ impl Printer {
     ) -> Result<()> {
         self.result.push_str(name);
         if memarg.memory != 0 {
-            write!(self.result, " (memory {})", memarg.memory)?;
+            self.result.push_str(" (memory ");
+            self.print_memory_idx(memarg.memory)?;
+            self.result.push_str(")");
         }
         if memarg.offset != 0 {
             write!(self.result, " offset={}", memarg.offset)?;
@@ -1499,20 +1556,29 @@ impl Printer {
         Ok(())
     }
 
-    fn print_blockty(&mut self, ty: &TypeOrFuncType) -> Result<()> {
+    fn print_blockty(&mut self, ty: &TypeOrFuncType, cur_depth: u32) -> Result<()> {
+        if let Some(name) = self
+            .state
+            .label_names
+            .get(&(self.state.func, self.state.label))
+        {
+            self.result.push_str(" ");
+            name.write(&mut self.result);
+        }
         match ty {
-            TypeOrFuncType::Type(Type::EmptyBlockType) => Ok(()),
+            TypeOrFuncType::Type(Type::EmptyBlockType) => {}
             TypeOrFuncType::Type(t) => {
                 self.result.push_str(" (result ");
                 self.print_valtype(*t)?;
                 self.result.push_str(")");
-                Ok(())
             }
             TypeOrFuncType::FuncType(idx) => {
                 self.print_functype_idx(*idx, false, None)?;
-                Ok(())
             }
         }
+        write!(self.result, "  ;; label = @{}", cur_depth)?;
+        self.state.label += 1;
+        Ok(())
     }
 
     fn print_exports(&mut self, data: ExportSectionReader) -> Result<()> {
@@ -1522,21 +1588,7 @@ impl Printer {
             self.start_group("export ");
             self.print_str(export.field)?;
             self.result.push_str(" ");
-            self.start_group("");
-            match export.kind {
-                ExternalKind::Function => {
-                    self.result.push_str("func ");
-                    self.print_func_idx(export.index)?;
-                }
-                ExternalKind::Table => write!(self.result, "table {}", export.index)?,
-                ExternalKind::Global => write!(self.result, "global {}", export.index)?,
-                ExternalKind::Memory => write!(self.result, "memory {}", export.index)?,
-                ExternalKind::Event => write!(self.result, "event {}", export.index)?,
-                ExternalKind::Module => write!(self.result, "module {}", export.index)?,
-                ExternalKind::Instance => write!(self.result, "instance {}", export.index)?,
-                ExternalKind::Type => write!(self.result, "type {}", export.index)?,
-            }
-            self.end_group(); // field
+            self.print_external(export.kind, export.index)?;
             self.end_group(); // export
         }
         return Ok(());
@@ -1549,13 +1601,25 @@ impl Printer {
                 self.result.push_str("func ");
                 self.print_func_idx(index)?;
             }
-            ExternalKind::Table => write!(self.result, "table {}", index)?,
-            ExternalKind::Global => write!(self.result, "global {}", index)?,
-            ExternalKind::Memory => write!(self.result, "memory {}", index)?,
-            ExternalKind::Event => write!(self.result, "event {}", index)?,
+            ExternalKind::Table => {
+                self.result.push_str("table ");
+                self.print_table_idx(index)?;
+            }
+            ExternalKind::Global => {
+                self.result.push_str("global ");
+                self.print_global_idx(index)?;
+            }
+            ExternalKind::Memory => {
+                self.result.push_str("memory ");
+                self.print_memory_idx(index)?;
+            }
+            ExternalKind::Tag => write!(self.result, "tag {}", index)?,
             ExternalKind::Module => write!(self.result, "module {}", index)?,
             ExternalKind::Instance => write!(self.result, "instance {}", index)?,
-            ExternalKind::Type => write!(self.result, "type {}", index)?,
+            ExternalKind::Type => {
+                self.result.push_str("type ");
+                self.print_type_idx(index)?;
+            }
         }
         self.result.push_str(")");
         Ok(())
@@ -1566,15 +1630,127 @@ impl Printer {
     ///
     /// This will either print `$foo` or `idx` as a raw integer.
     pub fn print_func_idx(&mut self, idx: u32) -> Result<()> {
-        match self.state.names.get(&idx) {
+        match self.state.function_names.get(&idx) {
             Some(name) => write!(self.result, "${}", name.identifier())?,
             None => write!(self.result, "{}", idx)?,
         }
         Ok(())
     }
 
+    fn print_cur_func_name(&mut self) -> Result<()> {
+        match self.state.function_names.get(&self.state.func) {
+            Some(name) => name.write(&mut self.result),
+            None => write!(self.result, "(;{};)", self.state.func)?,
+        }
+        Ok(())
+    }
+
+    fn print_global_idx(&mut self, idx: u32) -> Result<()> {
+        match self.state.global_names.get(&idx) {
+            Some(name) => write!(self.result, "${}", name.identifier())?,
+            None => write!(self.result, "{}", idx)?,
+        }
+        Ok(())
+    }
+
+    fn print_cur_global_name(&mut self) -> Result<()> {
+        match self.state.global_names.get(&self.state.global) {
+            Some(name) => name.write(&mut self.result),
+            None => write!(self.result, "(;{};)", self.state.global)?,
+        }
+        Ok(())
+    }
+
+    fn print_table_idx(&mut self, idx: u32) -> Result<()> {
+        match self.state.table_names.get(&idx) {
+            Some(name) => write!(self.result, "${}", name.identifier())?,
+            None => write!(self.result, "{}", idx)?,
+        }
+        Ok(())
+    }
+
+    fn print_cur_table_name(&mut self) -> Result<()> {
+        match self.state.table_names.get(&self.state.table) {
+            Some(name) => name.write(&mut self.result),
+            None => write!(self.result, "(;{};)", self.state.table)?,
+        }
+        Ok(())
+    }
+
+    fn print_memory_idx(&mut self, idx: u32) -> Result<()> {
+        match self.state.memory_names.get(&idx) {
+            Some(name) => write!(self.result, "${}", name.identifier())?,
+            None => write!(self.result, "{}", idx)?,
+        }
+        Ok(())
+    }
+
+    fn print_cur_memory_name(&mut self) -> Result<()> {
+        match self.state.memory_names.get(&self.state.memory) {
+            Some(name) => name.write(&mut self.result),
+            None => write!(self.result, "(;{};)", self.state.memory)?,
+        }
+        Ok(())
+    }
+
+    fn print_type_ref(&mut self, idx: u32) -> Result<()> {
+        self.result.push_str(" (type ");
+        self.print_type_idx(idx)?;
+        self.result.push_str(")");
+        Ok(())
+    }
+
+    fn print_type_idx(&mut self, idx: u32) -> Result<()> {
+        match self.state.type_names.get(&idx) {
+            Some(name) => write!(self.result, "${}", name.identifier())?,
+            None => write!(self.result, "{}", idx)?,
+        }
+        Ok(())
+    }
+
+    fn print_cur_type_name(&mut self) -> Result<()> {
+        let cur_idx = self.state.types.len() as u32;
+        match self.state.type_names.get(&cur_idx) {
+            Some(name) => name.write(&mut self.result),
+            None => write!(self.result, "(;{};)", cur_idx)?,
+        }
+        Ok(())
+    }
+
+    fn print_data_idx(&mut self, idx: u32) -> Result<()> {
+        match self.state.data_names.get(&idx) {
+            Some(name) => write!(self.result, "${}", name.identifier())?,
+            None => write!(self.result, "{}", idx)?,
+        }
+        Ok(())
+    }
+
+    fn print_data_name(&mut self, idx: u32) -> Result<()> {
+        match self.state.data_names.get(&idx) {
+            Some(name) => name.write(&mut self.result),
+            None => write!(self.result, "(;{};)", idx)?,
+        }
+        Ok(())
+    }
+
+    fn print_elem_idx(&mut self, idx: u32) -> Result<()> {
+        match self.state.element_names.get(&idx) {
+            Some(name) => write!(self.result, "${}", name.identifier())?,
+            None => write!(self.result, "{}", idx)?,
+        }
+        Ok(())
+    }
+
+    fn print_elem_name(&mut self, idx: u32) -> Result<()> {
+        match self.state.element_names.get(&idx) {
+            Some(name) => name.write(&mut self.result),
+            None => write!(self.result, "(;{};)", idx)?,
+        }
+        Ok(())
+    }
+
     fn print_local_idx(&mut self, func: u32, idx: u32) -> Result<()> {
-        match self.state.local_names.get(&func).and_then(|f| f.get(&idx)) {
+        match self.state.local_names.get(&(func, idx)) {
             Some(name) => write!(self.result, "${}", name.identifier())?,
             None => write!(self.result, "{}", idx)?,
         }
@@ -1585,8 +1761,8 @@ impl Printer {
         for (i, elem) in data.into_iter().enumerate() {
             let mut elem = elem?;
             self.newline();
-            self.start_group("elem");
-            write!(self.result, " (;{};)", i)?;
+            self.start_group("elem ");
+            self.print_elem_name(i as u32)?;
             match &mut elem.kind {
                 ElementKind::Passive => {}
                 ElementKind::Declared => write!(self.result, " declare")?,
@@ -1595,7 +1771,9 @@ impl Printer {
                     init_expr,
                 } => {
                     if *table_index != 0 {
-                        write!(self.result, " (table {})", table_index)?;
+                        self.result.push_str(" (table ");
+                        self.print_table_idx(*table_index)?;
+                        self.result.push_str(")");
                     }
                     self.result.push_str(" ");
                     self.print_init_expr(&init_expr)?;
@@ -1635,8 +1813,9 @@ impl Printer {
         for (i, data) in data.into_iter().enumerate() {
             let data = data?;
             self.newline();
-            self.start_group("data");
-            write!(self.result, " (;{};) ", i)?;
+            self.start_group("data ");
+            self.print_data_name(i as u32)?;
+            self.result.push_str(" ");
             match &data.kind {
                 DataKind::Passive => {}
                 DataKind::Active {
@@ -1644,7 +1823,9 @@ impl Printer {
                     init_expr,
                 } => {
                     if *memory_index != 0 {
-                        write!(self.result, " (memory {}) ", memory_index)?;
+                        self.result.push_str(" (memory ");
+                        self.print_memory_idx(*memory_index)?;
+                        self.result.push_str(")");
                     }
                     self.print_init_expr(&init_expr)?;
                     self.result.push_str(" ");
@@ -1699,7 +1880,7 @@ impl Printer {
                         ExternalKind::Function => self.start_group("func"),
                         ExternalKind::Table => self.start_group("table"),
                         ExternalKind::Memory => self.start_group("memory"),
-                        ExternalKind::Event => self.start_group("event"),
+                        ExternalKind::Tag => self.start_group("tag"),
                         ExternalKind::Global => self.start_group("global"),
                         ExternalKind::Instance => self.start_group("instance"),
                         ExternalKind::Module => self.start_group("module"),
@@ -1708,26 +1889,23 @@ impl Printer {
                     self.result.push_str(" ");
                     match kind {
                         ExternalKind::Function => {
-                            match self.state.names.get(&self.state.func) {
-                                Some(name) => write!(self.result, "${}", name.identifier())?,
-                                None => write!(self.result, "(;{};)", self.state.func)?,
-                            }
+                            self.print_cur_func_name()?;
                             self.state.func += 1;
                         }
                         ExternalKind::Table => {
-                            write!(self.result, "(;{};)", self.state.table)?;
+                            self.print_cur_table_name()?;
                             self.state.table += 1;
                         }
                         ExternalKind::Memory => {
-                            write!(self.result, "(;{};)", self.state.memory)?;
+                            self.print_cur_memory_name()?;
                             self.state.memory += 1;
                         }
-                        ExternalKind::Event => {
-                            write!(self.result, "(;{};)", self.state.event)?;
-                            self.state.event += 1;
+                        ExternalKind::Tag => {
+                            write!(self.result, "(;{};)", self.state.tag)?;
+                            self.state.tag += 1;
                         }
                         ExternalKind::Global => {
-                            write!(self.result, "(;{};)", self.state.global)?;
+                            self.print_cur_global_name()?;
                             self.state.global += 1;
                         }
                         ExternalKind::Instance => {
@@ -1739,7 +1917,7 @@ impl Printer {
                             self.state.module += 1;
                         }
                         ExternalKind::Type => {
-                            write!(self.result, "(;{};)", self.state.types.len())?;
+                            self.print_cur_type_name()?;
                             self.state.types.push(None);
                         }
                     }
@@ -1749,13 +1927,9 @@ impl Printer {
                     relative_depth,
                     index,
                 } => {
-                    write!(
-                        self.result,
-                        "outer {} {} (type (;{};))",
-                        relative_depth,
-                        index,
-                        self.state.types.len(),
-                    )?;
+                    write!(self.result, "outer {} {} (type ", relative_depth, index)?;
+                    self.print_cur_type_name()?;
+                    self.result.push_str(")");
                     self.state.types.push(None);
                 }
                 Alias::OuterModule {
